package org.greenstone.atea;

import java.io.*;

import org.apache.commons.csv.*;
import org.apache.log4j.Logger;

import com.google.gson.*;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

// For working with GeoJSON's Simple Features in Java
import mil.nga.sf.geojson.FeatureConverter;
import mil.nga.sf.geojson.Geometry;
import mil.nga.sf.geojson.MultiPoint;
import mil.nga.sf.geojson.Position;


/** Simple Features GeoJSON Java
 * https://ngageoint.github.io/simple-features-geojson-java/ - liks to API and more
 *
 * https://mvnrepository.com/artifact/mil.nga.sf/sf-geojson (https://github.com/ngageoint/simple-features-geojson-java/)
 *
 * Also need the basic data types used by the Geometry objects above: 
 * https://mvnrepository.com/artifact/mil.nga/sf (https://github.com/ngageoint/simple-features-java)
 *
 * Further helper jars needed (because of encountering the exception documented at
 * stackoverflow.com/questions/36278293/java-lang-classnotfoundexception-com-fasterxml-jackson-core-jsonprocessingexcep/36279872)
 *   https://mvnrepository.com/artifact/com.fasterxml.jackson.core/jackson-core/2.10.0
 *   https://mvnrepository.com/artifact/com.fasterxml.jackson.core/jackson-databind
 *   https://mvnrepository.com/artifact/com.fasterxml.jackson.core/jackson-annotations/2.10.0
 */

/**
 * This class needs the gson library, and now the sf(-2.02).jar and sf-geojson(-2.02).jar files too
 * to create and store Simple Features geo json objects with Java.
 * I copied the gson jar file from GS3.
 *
 * TO COMPILE:
 *    maori-lang-detection/src$
 *       javac -cp ".:../conf:../lib/*" org/greenstone/atea/CountryCodeCountsMapData.java
 *
 * TO RUN:
 *    maori-lang-detection/src$
 *       java -cp ".:../conf:../lib/*" org/greenstone/atea/CountryCodeCountsMapData ../mongodb-data/countrycodes.json ../mongodb-data/counts.json
 *
 */
public class CountryCodeCountsMapData {
    static Logger logger = Logger.getLogger(org.greenstone.atea.CountryCodeCountsMapData.class.getName());

    //Map<String, JsonObject> countryToJsonMap;
    JsonArray countryCodesJsonArray;
    JsonArray countryCountsJsonArray;

    // North-central Antarctica coords
    private final double ANTARCTICA_LNG = 57.0d;
    private final double ANTARCTICA_LAT = -70.0d;
    // For EU coords, spot in Atlantic Ocean close to western European coast.
    private final double EU_LNG = -20.0d;
    private final double EU_LAT = 50.0d;
    
    public CountryCodeCountsMapData(String countryCoordsJSONFile, String countryCountsJSONFile) throws Exception {
	
	// map of country codes to lat, lng json for that country code
	Map<String, JsonObject> countryToJsonMap = new HashMap<String, JsonObject>();

	// Parse json file of country codes and put into a JsonArray.
	// then put into map of each country code to its JsonObject.
	countryCodesJsonArray = parseJSONFile(countryCoordsJSONFile);
	for(JsonElement obj : countryCodesJsonArray) {
	    JsonObject countryCodeJson = obj.getAsJsonObject();
	    countryToJsonMap.put(countryCodeJson.get("country").getAsString(), countryCodeJson);
	}

	// Parse json file of country code counts
	// Then for each JsonObject in this file,
	// find a match on its country code in the map created above to get a country code JsonObject
	// Get the longitude and latitude of the JsonObject that matched that country code.
	// Add this lng,lat location information to the current JsonObject from the counts file.
	countryCountsJsonArray = parseJSONFile(countryCountsJSONFile);	

	for(JsonElement obj : countryCountsJsonArray) {
	    JsonObject json = obj.getAsJsonObject();
	    String countryCode = json.get("_id").getAsString();
	    int count = (int)json.get("count").getAsDouble();	    
	    
	    //logger.info("Got country code: " + countryCode);
	    //logger.info("   count: " + count);

	    // locate in countryCode map
	    JsonObject countryCodeJson = countryToJsonMap.get(countryCode);

	    if(countryCodeJson != null) {
		//logger.info("Found in map: " + countryCodeJson.toString());
		
		// for geojson, want longitude then latitude
		Double lng = countryCodeJson.get("longitude").getAsDouble();
		Double lat = countryCodeJson.get("latitude").getAsDouble();
		//logger.info("long: " + Double.toString(lng) + ", lat: " + Double.toString(lat));
		
		// let's add lat and lng fields to countryCounts object
		json.addProperty("lng", lng); // adds Number: https://javadoc.io/static/com.google.code.gson/gson/2.8.5/index.html?com/google/gson/Gson.html
		json.addProperty("lat", lat);
	    } else {
		logger.info("No geolocation info found for country code " + countryCode);
		if(countryCode.equals("EU")) {
		    //logger.info("Unlisted country code: EU");
		    // add lat and lng for Europe
		    json.addProperty("lng", EU_LNG); 
		    json.addProperty("lat", EU_LAT);		    
		}
		else if(countryCode.equals("UNKNOWN")) {
		    //logger.info("Unlisted country code: UNKNOWN");
		    // add lat and lng for Antarctica
		    json.addProperty("lng", ANTARCTICA_LNG); 
		    json.addProperty("lat", ANTARCTICA_LAT);
		} else {
		    logger.error("ERROR: entirely unknown country code: " + countryCode);
		}
	    }
	}

    }

    /** Convert mongodb tabular output of json records stored in the given file
     * into a JsonArray.
     */
    public JsonArray parseJSONFile(String filename) throws Exception {
	JsonArray jsonArray = null;
	// read into string
	try (
	     BufferedReader reader = new BufferedReader(new FileReader(filename));
	     ) {
	    StringBuilder str = //new StringBuilder();
		new StringBuilder("[");
	    String line;
	    while((line = reader.readLine()) != null) {
		line = line.replaceAll("/\\* [^\\/]* \\*/", "");
		str.append(line);
		if(line.endsWith("}")) {
		    str.append(",\n");
		}
	    }
	    // replace last comma with closing bracket
	    String fileContents = str.substring(0, str.length()-2) + "]";
	    
	    //System.err.println("Got file:\n" + fileContents);
	    
	    // https://stackoverflow.com/questions/2591098/how-to-parse-json-in-java
	    jsonArray = new JsonParser().parse(fileContents).getAsJsonArray();    
	    
	} catch(Exception e) {
	    throw e;
	}
	

	return jsonArray;
    }

    /**
     * Reading
     * https://www.here.xyz/api/concepts/geojsonbasics/
     * https://ngageoint.github.io/simple-features-geojson-java/docs/api/
     *
     * https://stackoverflow.com/questions/55621480/cant-access-coordinates-member-of-geojson-feature-collection
     *
     * Downloaded geojson simple features' jar file from maven, but it didn't work:
     * a more private version of MultiPoint.java is not included in the jar file (there's only
     * mil.nga.sf.geojson.MultiPoint  , whereas
     * mil.nga.sf.MultiPoint          is missing
     * 
     * This seems to have gone wrong at
     * https://github.com/ngageoint/simple-features-geojson-java/tree/master/src/main/java/mil/nga/sf
     * but the one at
     * https://github.com/ngageoint/simple-features-java/tree/master/src/main/java/mil/nga/sf
     * has it. So I've been trying to build that, but don't have the correct version of maven.
     */
    public Geometry toMultiPointGeoJson() {
	//System.err.println("toGeoJSON() is not yet implemented.");

	List<Position> points = new LinkedList<Position>();
	
	for(JsonElement obj : countryCountsJsonArray) {
	    JsonObject json = obj.getAsJsonObject();
	    Double lng = json.get("lng").getAsDouble();
	    Double lat = json.get("lat").getAsDouble();

	    Position point = new Position(lng, lat);
	    points.add(point);
	}
	
	Geometry multiPoint = new MultiPoint(points);
	
	return multiPoint;
    }


    public String writeMultiPointGeoJsonToFile(File folder) {
	final String filename = "multipoint.json";
	File outFile = new File(folder, filename);

	Geometry geometry = this.toMultiPointGeoJson();
	String multiPointGeojsonString = FeatureConverter.toStringValue(geometry);
	System.err.println("\nMap data as MultiPoint geometry:\n" + multiPointGeojsonString + "\n");
	try (
	     Writer writer = new BufferedWriter(new FileWriter(outFile));
	     ) {

	    // Some basic re-formatting for some immediate legibility
	    // But pasting the contents of the file (or the System.err output above)
	    // directly into http://geojson.tools/ will instantly reformat the json perfectly anyway.
	    multiPointGeojsonString = multiPointGeojsonString.replace("[[", "\n[\n\t[");
	    multiPointGeojsonString = multiPointGeojsonString.replace("],[", "],\n\t[");
	    multiPointGeojsonString = multiPointGeojsonString.replace("]]", "]\n]");
	    
	    writer.write(multiPointGeojsonString + "\n");
	} catch(Exception e) {
	    logger.error("Unable to write multipoint geojson:\n**********************");
	    logger.error(multiPointGeojsonString);
	    logger.error("**********************\ninto file " + outFile.getAbsolutePath());
	    logger.error(e.getMessage(), e);
	}

	return outFile.getAbsolutePath();
	
    }
    
    
    // Unfinished and unused
    public void parseCSVFile(String filename) throws Exception {
	File csvData = new File(filename);
	CSVParser parser = CSVParser.parse(csvData, java.nio.charset.Charset.forName("US-ASCII"), CSVFormat.RFC4180);
	for (CSVRecord csvRecord : parser) {
	    logger.info("Got record: " + csvRecord.toString());
	}
    }
    
    public static void printUsage() {
	System.err.println("CountryCodeCountsMapData countrycodes.json counts.json");
    }
    
    public static void main(String args[]) {
	if(args.length != 2) {
	    printUsage();
	    System.exit(-1);
	}
	
	try {
	    File countsFile = new File(args[1]);
	    File parentFolder = countsFile.getParentFile().getCanonicalFile(); // canonical resolves any .. and . in path
		
	    CountryCodeCountsMapData mapData = new CountryCodeCountsMapData(args[0], args[1]);

	    //Geometry geometry = mapData.toMultiPointGeoJSON();
	    //String multiPointGeojsonString = FeatureConverter.toStringValue(geometry);
	    //System.err.println("geometry: " + multiPointGeojsonString);

	    String outFileName = mapData.writeMultiPointGeoJsonToFile(parentFolder);
	    System.err.println("***********\nWrote mapdata to file " + outFileName);
	} catch(Exception e) {
	    logger.error(e.getMessage(), e);
	}
    }
}
